@@ -34,6 +34,7 @@
 
 #include <libebook/e-contact.h>
 #include <libedataserver/e-data-server-util.h>
+#include <libedataserver/e-sexp.h>
 #include <camel/camel.h>
 
 #include <e-mapi-operation-queue.h>
@@ -57,9 +58,14 @@ struct _EBookBackendMAPIPrivate
 
 	GThread *update_cache_thread;
 	GCancellable *update_cache;
+	time_t last_update_cache;
 
 	EBookBackendSqliteDB *db;
 	GHashTable *running_book_views;
+
+	guint32 last_server_contact_count;
+	time_t last_modify_time;
+	gboolean server_dirty;
 };
 
 #define ELEMENT_TYPE_MASK   0xF /* mask where the real type of the element is stored */
@@ -71,8 +77,8 @@ struct _EBookBackendMAPIPrivate
 
 static const struct field_element_mapping {
 		EContactField field_id;
-		gint element_type;
-		gint mapi_id;
+		uint32_t element_type;
+		uint32_t mapi_id;
 		gint contact_type;
 	} mappings [] = {
 
@@ -123,60 +129,6 @@ static const struct field_element_mapping {
 	/* { E_CONTACT_CATEGORIES, } */
 	};
 
-static gboolean
-ebbm_get_cache_time (EBookBackendMAPI *ebma, glong *cache_seconds)
-{
-	GError *error = NULL;
-	GTimeVal tv = { 0 };
-	gchar *last_update;
-	gboolean ret = TRUE;
-
-	g_return_val_if_fail (ebma != NULL, FALSE);
-	g_return_val_if_fail (ebma->priv != NULL, FALSE);
-	g_return_val_if_fail (ebma->priv->db != NULL, FALSE);
-	g_return_val_if_fail (cache_seconds != NULL, FALSE);
-
-	last_update = e_book_backend_sqlitedb_get_sync_data (ebma->priv->db, EMA_EBB_CACHE_FOLDERID, &error);
-	ret = !error && last_update && g_time_val_from_iso8601 (last_update, &tv);
-	if (error)
-		g_error_free (error);
-	g_free (last_update);
-
-	if (ret)
-		*cache_seconds = tv.tv_sec;
-
-	return ret;
-}
-
-static void
-ebbm_set_cache_time (EBookBackendMAPI *ebma, glong cache_seconds)
-{
-	GError *error = NULL;
-	gchar *iso_time = NULL;
-
-	g_return_if_fail (ebma != NULL);
-	g_return_if_fail (ebma->priv != NULL);
-	g_return_if_fail (ebma->priv->db != NULL);
-
-	if (cache_seconds > 0) {
-		GTimeVal tv = { 0 };
-
-		tv.tv_sec = cache_seconds;
-		iso_time = g_time_val_to_iso8601 (&tv);
-	}
-
-	e_book_backend_sqlitedb_set_sync_data (ebma->priv->db,
-					       EMA_EBB_CACHE_FOLDERID,
-					       iso_time ? iso_time : "0",
-					       &error);
-	g_free (iso_time);
-
-	if (error) {
-		e_mapi_debug_print ("%s: Failed to set value: %s", G_STRFUNC, error->message);
-		g_error_free (error);
-	}
-}
-
 static gboolean
 pick_view_cb (EDataBookView *view, gpointer user_data)
 {
@@ -213,12 +165,6 @@ complete_views (EBookBackendMAPI *ebma)
 	e_book_backend_foreach_view (E_BOOK_BACKEND (ebma), complete_view_cb, NULL);
 }
 
-struct FetchContactsData
-{
-	glong last_notification;
-	glong last_modification;
-};
-
 static void
 ebbm_notify_connection_status (EBookBackendMAPI *ebma, gboolean is_online)
 {
@@ -235,10 +181,14 @@ ebbm_notify_connection_status (EBookBackendMAPI *ebma, gboolean is_online)
 }
 
 static void
-ebbm_fetch_contacts (EBookBackendMAPI *ebma, BuildRestrictionsCB build_rs_cb, gpointer build_rs_cb_data, EDataBookView *book_view, glong *last_modification_secs, GError **error)
+ebbm_transfer_contacts (EBookBackendMAPI *ebma,
+			const GSList *uids,
+			EDataBookView *book_view,
+			GCancellable *cancellable,
+			GError **error)
 {
 	EBookBackendMAPIClass *ebmac;
-	struct FetchContactsData notify_data = { 0 };
+	glong last_notification = 0;
 
 	g_return_if_fail (ebma != NULL);
 	g_return_if_fail (ebma->priv != NULL);
@@ -246,12 +196,9 @@ ebbm_fetch_contacts (EBookBackendMAPI *ebma, BuildRestrictionsCB build_rs_cb, gp
 
 	ebmac = E_BOOK_BACKEND_MAPI_GET_CLASS (ebma);
 	g_return_if_fail (ebmac != NULL);
-	g_return_if_fail (ebmac->op_fetch_contacts != NULL);
+	g_return_if_fail (ebmac->op_transfer_contacts != NULL);
 
-	ebmac->op_fetch_contacts (ebma, build_rs_cb, build_rs_cb_data, book_view, &notify_data, error);
-
-	if (last_modification_secs && *last_modification_secs < notify_data.last_modification)
-		*last_modification_secs = notify_data.last_modification;
+	ebmac->op_transfer_contacts (ebma, uids, book_view, &last_notification, cancellable, error);
 }
 
 static gboolean
@@ -266,59 +213,16 @@ unref_backend_idle_cb (gpointer data)
 	return FALSE;
 }
 
-static gboolean
-ebbm_build_cache_update_restriction (EMapiConnection *conn,
-				     mapi_id_t fid,
-				     TALLOC_CTX *mem_ctx,
-				     struct mapi_SRestriction **restrictions,
-				     gpointer user_data,
-				     GCancellable *cancellable,
-				     GError **perror)
-{
-	EBookBackendMAPI *ebma = user_data;
-	struct mapi_SRestriction *restriction;
-	EBookBackendMAPIPrivate *priv;
-	struct SPropValue sprop;
-	struct timeval t = { 0 };
-	glong last_update_secs = 0;
-
-	g_return_val_if_fail (ebma != NULL, FALSE);
-	g_return_val_if_fail (mem_ctx != NULL, FALSE);
-	g_return_val_if_fail (E_IS_BOOK_BACKEND_MAPI (ebma), FALSE);
-	g_return_val_if_fail (restrictions != NULL, FALSE);
-
-	priv = ebma->priv;
-	g_return_val_if_fail (priv != NULL, FALSE);
-	g_return_val_if_fail (priv->db != NULL, FALSE);
-
-	if (!ebbm_get_cache_time (ebma, &last_update_secs) || last_update_secs <= 0)
-		return TRUE;
-
-	restriction = talloc_zero (mem_ctx, struct mapi_SRestriction);
-	g_return_val_if_fail (restriction != NULL, FALSE);
-
-	restriction->rt = RES_PROPERTY;
-	restriction->res.resProperty.relop = RELOP_GE;
-	restriction->res.resProperty.ulPropTag = PR_LAST_MODIFICATION_TIME;
-
-	t.tv_sec = last_update_secs;
-	t.tv_usec = 0;
-
-	set_SPropValue_proptag_date_timeval (&sprop, PR_LAST_MODIFICATION_TIME, &t);
-	cast_mapi_SPropValue (mem_ctx, &(restriction->res.resProperty.lpProp), &sprop);
-
-	*restrictions = restriction;
-
-	return TRUE;
-}
-
 static gpointer
 ebbm_update_cache_cb (gpointer data)
 {
 	EBookBackendMAPI *ebma = (EBookBackendMAPI *) data;
 	EBookBackendMAPIPrivate *priv;
 	EBookBackendMAPIClass *ebmac;
-	glong last_modification_secs = 0;
+	guint32 server_stored_contacts = 0;
+	time_t restr_tt = 0;
+	gboolean partial_update = FALSE;
+	GCancellable *cancellable;
 	GError *error = NULL;
 
 	g_return_val_if_fail (ebma != NULL, NULL);
@@ -332,47 +236,83 @@ ebbm_update_cache_cb (gpointer data)
 	ebmac = E_BOOK_BACKEND_MAPI_GET_CLASS (ebma);
 	g_return_val_if_fail (ebmac != NULL, NULL);
 
-	g_cancellable_reset (priv->update_cache);
 
-	if (!g_cancellable_is_cancelled (priv->update_cache) && ebmac->op_fetch_contacts) {
-		/* get time stored in a cache, to always use the latest last modification time */
-		if (!ebbm_get_cache_time (ebma, &last_modification_secs))
-			last_modification_secs = 0;
+	cancellable = priv->update_cache;
+	g_cancellable_reset (cancellable);
 
-		ebbm_fetch_contacts (ebma, ebbm_build_cache_update_restriction, ebma, NULL, &last_modification_secs, &error);
-		e_book_backend_sqlitedb_set_is_populated (priv->db, EMA_EBB_CACHE_FOLDERID, error != NULL, NULL);
-	}
+	do {
+		GHashTable *local_known_uids, *server_known_uids;
 
-	if (!error && !g_cancellable_is_cancelled (priv->update_cache) && ebmac->op_fetch_known_uids) {
-		GHashTable *uids = g_hash_table_new_full (g_str_hash, g_str_equal, (GDestroyNotify) g_free, NULL);
+		priv->server_dirty = FALSE;
 
-		ebmac->op_fetch_known_uids (ebma, priv->update_cache, uids, &error);
+		local_known_uids = e_book_backend_sqlitedb_get_uids_and_rev (priv->db, EMA_EBB_CACHE_FOLDERID, &error);
+		server_known_uids = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
-		if (!error && !g_cancellable_is_cancelled (priv->update_cache)) {
-			GSList *cache_keys, *c;
-
-			cache_keys = e_book_backend_sqlitedb_search_uids (priv->db, EMA_EBB_CACHE_FOLDERID, NULL, NULL, NULL);
-
-			for (c = cache_keys; c; c = c->next) {
-				const gchar *uid = c->data;
-
-				if (!uid || g_hash_table_lookup (uids, uid))
-					continue;
-
-				e_book_backend_mapi_notify_contact_removed (ebma, uid);
-			}
-
-			ebbm_set_cache_time (ebma, last_modification_secs);
-
-			g_slist_foreach (cache_keys, (GFunc) g_free, NULL);
-			g_slist_free (cache_keys);
+		if (!error && !g_cancellable_is_cancelled (cancellable) && ebmac->op_get_contacts_count) {
+			ebmac->op_get_contacts_count (ebma, &server_stored_contacts, cancellable, &error);
 		}
 
-		g_hash_table_destroy (uids);
-	}
+		if (!error && !g_cancellable_is_cancelled (cancellable) && ebmac->op_list_known_uids) {
+			struct ListKnownUidsData lku;
 
-	if (error)
-		g_error_free (error);
+			restr_tt = priv->last_modify_time && server_stored_contacts == g_hash_table_size (local_known_uids) ? priv->last_modify_time + 1 : 0;
+			partial_update = restr_tt > 0;
+
+			lku.uid_to_rev = server_known_uids;
+			lku.latest_last_modify = priv->last_modify_time;
+
+			ebmac->op_list_known_uids (ebma, partial_update ? e_mapi_utils_build_last_modify_restriction : NULL, &restr_tt, &lku, cancellable, &error);
+
+			restr_tt = lku.latest_last_modify;
+		}
+
+		if (!error && !g_cancellable_is_cancelled (cancellable) && ebmac->op_transfer_contacts && local_known_uids) {
+			GSList *uids = NULL;
+			GHashTableIter iter;
+			gpointer key, value;
+
+			g_hash_table_iter_init (&iter, server_known_uids);
+			while (g_hash_table_iter_next (&iter, &key, &value)) {
+				const gchar *uid = key, *rev = value, *local_rev;
+
+				local_rev = g_hash_table_lookup (local_known_uids, uid);
+				if (g_strcmp0 (local_rev, rev) != 0) {
+					uids = g_slist_prepend (uids, (gpointer) uid);
+				}
+
+				g_hash_table_remove (local_known_uids, uid);
+			}
+
+			if (uids)
+				ebbm_transfer_contacts (ebma, uids, NULL, cancellable, &error);
+
+			if (!error && !g_cancellable_is_cancelled (cancellable) && !partial_update) {
+				g_hash_table_iter_init (&iter, local_known_uids);
+				while (g_hash_table_iter_next (&iter, &key, &value)) {
+					const gchar *uid = key;
+
+					if (!uid)
+						continue;
+
+					e_book_backend_mapi_notify_contact_removed (ebma, uid);
+				}
+			}
+
+			priv->last_server_contact_count = server_stored_contacts;
+			priv->last_modify_time = restr_tt;
+
+			/* has borrowed data from server_known_uids */
+			g_slist_free (uids);
+		}
+
+		priv->last_update_cache = time(NULL);
+
+		g_hash_table_destroy (server_known_uids);
+		if (local_known_uids)
+			g_hash_table_destroy (local_known_uids);
+	} while (!error && priv->server_dirty && !g_cancellable_is_cancelled (cancellable));
+
+	g_clear_error (&error);
 
 	complete_views (ebma);
 
@@ -385,6 +325,36 @@ ebbm_update_cache_cb (gpointer data)
 	return NULL;
 }
 
+static void
+ebbm_maybe_invoke_cache_update (EBookBackendMAPI *ebma)
+{
+	EBookBackendMAPIPrivate *priv;
+
+	g_return_if_fail (ebma != NULL);
+	g_return_if_fail (ebma->priv != NULL);
+
+	priv = ebma->priv;
+
+	if (priv->update_cache_thread) {
+		if (!g_cancellable_is_cancelled (priv->update_cache))
+			return;
+
+		g_thread_join (priv->update_cache_thread);
+		priv->update_cache_thread = NULL;
+	}
+
+	/* do not update more often than each 10 minutes */
+	if (time (NULL) - priv->last_update_cache >= 60 * 10) {
+		g_object_ref (ebma);
+
+		g_cancellable_reset (priv->update_cache);
+		priv->server_dirty = FALSE;
+		priv->update_cache_thread = g_thread_create (ebbm_update_cache_cb, ebma, TRUE, NULL);
+		if (!priv->update_cache_thread)
+			g_object_unref (ebma);
+	}
+}
+
 static void
 ebbm_connect_user (EBookBackendMAPI *ebma, GCancellable *cancellable, const gchar *password, GError **error)
 {
@@ -445,11 +415,7 @@ ebbm_connect_user (EBookBackendMAPI *ebma, GCancellable *cancellable, const gcha
 		ebbm_notify_connection_status (ebma, TRUE);
 
 		if (!g_cancellable_is_cancelled (cancellable) /* && priv->marked_for_offline */) {
-			g_object_ref (ebma);
-
-			priv->update_cache_thread = g_thread_create (ebbm_update_cache_cb, ebma, TRUE, NULL);
-			if (!priv->update_cache_thread)
-				g_object_unref (ebma);
+			ebbm_maybe_invoke_cache_update (ebma);
 		}
 	}
 }
@@ -745,19 +711,9 @@ ebbm_book_view_thread (gpointer data)
 
 		ebmac = E_BOOK_BACKEND_MAPI_GET_CLASS (bvtd->ebma);
 		if (ebmac && ebmac->op_book_view_thread)
-			ebmac->op_book_view_thread (bvtd->ebma, bvtd->book_view, &error);
+			ebmac->op_book_view_thread (bvtd->ebma, bvtd->book_view, priv->update_cache, &error);
 
-		if (!error && !e_book_backend_sqlitedb_get_is_populated (priv->db, EMA_EBB_CACHE_FOLDERID, &error)) {
-			/* todo: create restriction based on the book_view */
-			if (!error) {
-				g_cancellable_reset (priv->update_cache);
-				ebbm_fetch_contacts (bvtd->ebma, NULL, NULL, bvtd->book_view, NULL, &error);
-				g_cancellable_cancel (priv->update_cache);
-			}
-
-			if (!error)
-				e_book_backend_sqlitedb_set_is_populated (priv->db, EMA_EBB_CACHE_FOLDERID, TRUE, &error);
-		}
+		ebbm_maybe_invoke_cache_update (bvtd->ebma);
 	}
 
 	if (error && g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
@@ -891,7 +847,7 @@ ebbm_operation_cb (OperationBase *op, gboolean cancelled, EBookBackend *backend)
 				const GSList *l;
 
 				for (l = added_contacts; l; l = l->next) {
-					e_book_backend_mapi_notify_contact_update (ebma, NULL, E_CONTACT (l->data), NULL, -1, -1, NULL);
+					e_book_backend_mapi_notify_contact_update (ebma, NULL, E_CONTACT (l->data), -1, -1, NULL);
 				}
 			}
 
@@ -947,7 +903,7 @@ ebbm_operation_cb (OperationBase *op, gboolean cancelled, EBookBackend *backend)
 				const GSList *l;
 
 				for (l = modified_contacts; l; l = l->next) {
-					e_book_backend_mapi_notify_contact_update (ebma, NULL, E_CONTACT (l->data), NULL, -1, -1, NULL);
+					e_book_backend_mapi_notify_contact_update (ebma, NULL, E_CONTACT (l->data), -1, -1, NULL);
 				}
 			}
 
@@ -1317,6 +1273,10 @@ e_book_backend_mapi_init (EBookBackendMAPI *ebma)
 
 	ebma->priv->update_cache = g_cancellable_new ();
 	ebma->priv->update_cache_thread = NULL;
+	ebma->priv->last_update_cache = 0;
+	ebma->priv->last_server_contact_count = 0;
+	ebma->priv->last_modify_time = 0;
+	ebma->priv->server_dirty = FALSE;
 
 	g_signal_connect (
 		ebma, "notify::online",
@@ -1371,30 +1331,31 @@ e_book_backend_mapi_class_init (EBookBackendMAPIClass *klass)
 
 	g_type_class_add_private (klass, sizeof (EBookBackendMAPIPrivate));
 
-	object_class->dispose                     = ebbm_dispose;
+	object_class->dispose			= ebbm_dispose;
 
-	backend_class->open			  = ebbm_op_open;
-	backend_class->remove			  = ebbm_op_remove;
-	backend_class->create_contacts		  = ebbm_op_create_contacts;
-	backend_class->remove_contacts		  = ebbm_op_remove_contacts;
-	backend_class->modify_contacts		  = ebbm_op_modify_contacts;
-	backend_class->get_contact                = ebbm_op_get_contact;
-	backend_class->get_contact_list           = ebbm_op_get_contact_list;
-	backend_class->start_book_view            = ebbm_op_start_book_view;
-	backend_class->stop_book_view             = ebbm_op_stop_book_view;
-	backend_class->authenticate_user          = ebbm_op_authenticate_user;
-	backend_class->get_backend_property	  = ebbm_op_get_backend_property;
-	klass->op_open				  = ebbm_open;
-	klass->op_remove                          = ebbm_remove;
-	klass->op_authenticate_user               = ebbm_authenticate_user;
-	klass->op_get_contact                     = ebbm_get_contact;
-	klass->op_get_contact_list                = ebbm_get_contact_list;
+	backend_class->open			= ebbm_op_open;
+	backend_class->remove			= ebbm_op_remove;
+	backend_class->create_contacts		= ebbm_op_create_contacts;
+	backend_class->remove_contacts		= ebbm_op_remove_contacts;
+	backend_class->modify_contacts		= ebbm_op_modify_contacts;
+	backend_class->get_contact		= ebbm_op_get_contact;
+	backend_class->get_contact_list		= ebbm_op_get_contact_list;
+	backend_class->start_book_view		= ebbm_op_start_book_view;
+	backend_class->stop_book_view		= ebbm_op_stop_book_view;
+	backend_class->authenticate_user	= ebbm_op_authenticate_user;
+	backend_class->get_backend_property	= ebbm_op_get_backend_property;
+	klass->op_open				= ebbm_open;
+	klass->op_remove			= ebbm_remove;
+	klass->op_authenticate_user		= ebbm_authenticate_user;
+	klass->op_get_contact			= ebbm_get_contact;
+	klass->op_get_contact_list		= ebbm_get_contact_list;
 
-	klass->op_connection_status_changed       = NULL;
-	klass->op_get_status_message              = NULL;
-	klass->op_book_view_thread                = NULL;
-	klass->op_fetch_contacts                  = NULL;
-	klass->op_fetch_known_uids                = NULL;
+	klass->op_connection_status_changed	= NULL;
+	klass->op_get_status_message		= NULL;
+	klass->op_book_view_thread		= NULL;
+	klass->op_get_contacts_count		= NULL;
+	klass->op_list_known_uids		= NULL;
+	klass->op_transfer_contacts		= NULL;
 }
 
 gboolean
@@ -1529,14 +1490,14 @@ get_current_time_ms (void)
 	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
 }
 
-/* called from op_fetch_contacts - book_view and notify_contact_data are taken from there;
-   notify_contact_data is a pointer to FetchContactsData, if not NULL;
+/* called from op_transfer_contacts - book_view and notify_contact_data are taken from there;
+   notify_contact_data is a pointer to glong last_notification, if not NULL;
    returns whether can continue with fetching */
 gboolean
-e_book_backend_mapi_notify_contact_update (EBookBackendMAPI *ebma, EDataBookView *pbook_view, EContact *contact, const struct timeval *pr_last_modification_time, gint index, gint total, gpointer notify_contact_data)
+e_book_backend_mapi_notify_contact_update (EBookBackendMAPI *ebma, EDataBookView *pbook_view, EContact *contact, gint index, gint total, gpointer notify_contact_data)
 {
 	EBookBackendMAPIPrivate *priv;
-	struct FetchContactsData *fcd = notify_contact_data;
+	glong *last_notification = notify_contact_data;
 	EDataBookView *book_view = pbook_view;
 	GError *error = NULL;
 
@@ -1553,13 +1514,13 @@ e_book_backend_mapi_notify_contact_update (EBookBackendMAPI *ebma, EDataBookView
 		book_view = ebbm_pick_book_view (ebma);
 
 	if (book_view) {
-		guint32 current_time;
+		glong current_time;
 
 		if (!e_book_backend_mapi_book_view_is_running (ebma, book_view))
 			return FALSE;
 
 		current_time = get_current_time_ms ();
-		if (index > 0 && fcd && current_time - fcd->last_notification > 333) {
+		if (index > 0 && last_notification && current_time - *last_notification > 333) {
 			gchar *status_msg = NULL;
 			EBookBackendMAPIClass *ebmac = E_BOOK_BACKEND_MAPI_GET_CLASS (ebma);
 
@@ -1571,7 +1532,7 @@ e_book_backend_mapi_notify_contact_update (EBookBackendMAPI *ebma, EDataBookView
 
 			g_free (status_msg);
 
-			fcd->last_notification = current_time;
+			*last_notification = current_time;
 		}
 	}
 
@@ -1583,14 +1544,11 @@ e_book_backend_mapi_notify_contact_update (EBookBackendMAPI *ebma, EDataBookView
 					     FALSE, &error);
 	if (!error) {
 		e_book_backend_notify_update (E_BOOK_BACKEND (ebma), contact);
-
-		if (fcd && pr_last_modification_time) {
-			if (fcd->last_modification < pr_last_modification_time->tv_sec)
-				fcd->last_modification = pr_last_modification_time->tv_sec;
-		}
 		return TRUE;
 	}
+
 	g_error_free (error);
+
 	return FALSE;
 }
 
@@ -1642,6 +1600,19 @@ e_book_backend_mapi_cache_get (EBookBackendMAPI *ebma, const gchar *key)
 	return e_book_backend_sqlitedb_get_key_value (ebma->priv->db, EMA_EBB_CACHE_FOLDERID, key, NULL);
 }
 
+void
+e_book_backend_mapi_refresh_cache (EBookBackendMAPI *ebma)
+{
+	g_return_if_fail (ebma != NULL);
+	g_return_if_fail (E_IS_BOOK_BACKEND_MAPI (ebma));
+
+	ebma->priv->last_update_cache = 0;
+	ebma->priv->last_modify_time = 0;
+	ebma->priv->server_dirty = TRUE;
+
+	ebbm_maybe_invoke_cache_update (ebma);
+}
+
 /* utility functions/macros */
 
 /* 'data' is one of GET_ALL_KNOWN_IDS or GET_UIDS_ONLY */
@@ -1708,9 +1679,10 @@ mapi_book_utils_get_prop_list (EMapiConnection *conn,
 	};
 
 	static uint32_t uids_only_ids[] = {
-		PR_FID,
-		PR_MID,
-		PR_EMAIL_ADDRESS_UNICODE
+		PidTagFolderId,
+		PidTagMid,
+		PidTagLastModificationTime,
+		PidTagEmailAddress
 	};
 
 	/* do not make this array static, the function modifies it on run */
@@ -1785,9 +1757,21 @@ mapi_book_utils_contact_from_props (EMapiConnection *conn, mapi_id_t fid, const
 
 	if (g_str_equal (get_str_proptag (PR_MESSAGE_CLASS), IPM_DISTLIST)) {
 		const struct mapi_SBinaryArray *members, *members_dlist;
+		const struct FILETIME *last_modification;
 		GSList *attrs = NULL, *a;
 		gint i;
 
+		last_modification = get_proptag (PidTagLastModificationTime);
+		if (last_modification) {
+			gchar *buff = NULL;
+
+			buff = mapi_book_utils_timet_to_string (e_mapi_util_filetime_to_time_t (last_modification));
+			if (buff)
+				e_contact_set (contact, E_CONTACT_REV, buff);
+
+			g_free (buff);
+		}
+
 		/* it's a contact list/distribution list, fetch members and return it */
 		e_contact_set (contact, E_CONTACT_IS_LIST, GINT_TO_POINTER (TRUE));
 		/* we do not support this option, same as GroupWise */
@@ -1879,13 +1863,8 @@ mapi_book_utils_contact_from_props (EMapiConnection *conn, mapi_id_t fid, const
 			if (value && mappings[i].element_type == PT_SYSTIME) {
 				const struct FILETIME *t = value;
 				gchar *buff = NULL;
-				GTimeVal tv;
-
-				tv.tv_sec = e_mapi_util_filetime_to_time_t (t);
-				tv.tv_usec = 0;
-				
-				buff = g_time_val_to_iso8601 (&tv);
 
+				buff = mapi_book_utils_timet_to_string (e_mapi_util_filetime_to_time_t (t));
 				if (buff)
 					e_contact_set (contact, mappings[i].field_id, buff);
 
@@ -1982,6 +1961,11 @@ mapi_error_to_edb_error (GError **perror, const GError *mapi_error, EDataBookSta
 	if (!perror)
 		return;
 
+	if (g_error_matches (mapi_error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
+		g_propagate_error (perror, g_error_copy (mapi_error));
+		return;
+	}
+
 	if (code == E_DATA_BOOK_STATUS_OTHER_ERROR && mapi_error && mapi_error->domain == E_MAPI_ERROR) {
 		/* Change error to more accurate only with OTHER_ERROR */
 		switch (mapi_error->code) {
@@ -2006,3 +1990,421 @@ mapi_error_to_edb_error (GError **perror, const GError *mapi_error, EDataBookSta
 
 	g_free (err_msg);
 }
+
+gchar *
+mapi_book_utils_timet_to_string (time_t tt)
+{
+	GTimeVal tv;
+
+	tv.tv_sec = tt;
+	tv.tv_usec = 0;
+
+	return g_time_val_to_iso8601 (&tv);
+}
+
+struct EMapiSExpParserData
+{
+	TALLOC_CTX *mem_ctx;
+	/* parser results in ints, indexes to res_parts */
+	GPtrArray *res_parts;
+};
+
+static ESExpResult *
+term_eval_and (struct _ESExp *f,
+	       gint argc,
+	       struct _ESExpResult **argv,
+	       gpointer user_data)
+{
+	struct EMapiSExpParserData *esp = user_data;
+	ESExpResult *r;
+	gint ii, jj, valid = 0;
+
+	r = e_sexp_result_new (f, ESEXP_RES_INT);
+	r->value.number = -1;
+
+	for (ii = 0; ii < argc; ii++) {
+		if (argv[ii]->type == ESEXP_RES_INT &&
+		    argv[ii]->value.number >= 0 && 
+		    argv[ii]->value.number < esp->res_parts->len) {
+			jj = argv[ii]->value.number;
+			valid++;
+		}
+	}
+
+	if (valid == 1) {
+		r->value.number = jj;
+	} else if (valid > 0) {
+		struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+		g_return_val_if_fail (res != NULL, NULL);
+
+		res->rt = RES_AND;
+		res->res.resAnd.cRes = valid;
+		res->res.resAnd.res = talloc_zero_array (esp->mem_ctx, struct mapi_SRestriction_and, res->res.resAnd.cRes + 1);
+
+		jj = 0;
+
+		for (ii = 0; ii < argc; ii++) {
+			if (argv[ii]->type == ESEXP_RES_INT &&
+			    argv[ii]->value.number >= 0 && 
+			    argv[ii]->value.number < esp->res_parts->len) {
+				struct mapi_SRestriction *subres = g_ptr_array_index (esp->res_parts, argv[ii]->value.number);
+
+				res->res.resAnd.res[jj].rt = subres->rt;
+				res->res.resAnd.res[jj].res = subres->res;
+
+				jj++;
+			}
+		}
+
+		g_ptr_array_add (esp->res_parts, res);
+		r->value.number = esp->res_parts->len - 1;
+	}
+
+	return r;
+}
+
+static ESExpResult *
+term_eval_or (struct _ESExp *f,
+	      gint argc,
+	      struct _ESExpResult **argv,
+	      gpointer user_data)
+{
+	struct EMapiSExpParserData *esp = user_data;
+	ESExpResult *r;
+	gint ii, jj = -1, valid = 0;
+
+	r = e_sexp_result_new (f, ESEXP_RES_INT);
+	r->value.number = -1;
+
+	for (ii = 0; ii < argc; ii++) {
+		if (argv[ii]->type == ESEXP_RES_INT &&
+		    argv[ii]->value.number >= 0 && 
+		    argv[ii]->value.number < esp->res_parts->len) {
+			jj = argv[ii]->value.number;
+			valid++;
+		    }
+	}
+
+	if (valid == 1) {
+		r->value.number = jj;
+	} else if (valid > 0) {
+		struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+		g_return_val_if_fail (res != NULL, NULL);
+
+		res->rt = RES_OR;
+		res->res.resOr.cRes = valid;
+		res->res.resOr.res = talloc_zero_array (esp->mem_ctx, struct mapi_SRestriction_or, res->res.resOr.cRes + 1);
+
+		jj = 0;
+
+		for (ii = 0; ii < argc; ii++) {
+			if (argv[ii]->type == ESEXP_RES_INT &&
+			    argv[ii]->value.number >= 0 && 
+			    argv[ii]->value.number < esp->res_parts->len) {
+				struct mapi_SRestriction *subres = g_ptr_array_index (esp->res_parts, argv[ii]->value.number);
+
+				res->res.resOr.res[jj].rt = subres->rt;
+				res->res.resOr.res[jj].res = subres->res;
+
+				jj++;
+			}
+		}
+
+		g_ptr_array_add (esp->res_parts, res);
+		r->value.number = esp->res_parts->len - 1;
+	}
+
+	return r;
+}
+
+static ESExpResult *
+term_eval_not (struct _ESExp *f,
+	       gint argc,
+	       struct _ESExpResult **argv,
+	       gpointer user_data)
+{
+	ESExpResult *r;
+
+	r = e_sexp_result_new (f, ESEXP_RES_INT);
+	r->value.number = -1;
+
+	#ifdef HAVE_RES_NOT_SUPPORTED
+	if (argc == 1 && argv[0]->type == ESEXP_RES_INT) {
+		struct EMapiSExpParserData *esp = user_data;
+		gint idx = argv[0]->value.number;
+
+		if (esp && idx >= 0 && idx < esp->res_parts->len) {
+			struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+			g_return_val_if_fail (res != NULL, NULL);
+
+			res->rt = RES_NOT;
+			res->res.resNot.res = g_ptr_array_index (esp->res_parts, idx);
+
+			g_ptr_array_add (esp->res_parts, res);
+			r->value.number = esp->res_parts->len - 1;
+		}
+	}
+	#endif
+
+	return r;
+}
+
+static uint32_t
+get_proptag_from_field_name (const gchar *field_name, gboolean is_contact_field)
+{
+	EContactField cfid;
+	gint ii;
+
+	if (is_contact_field)
+		cfid = e_contact_field_id (field_name);
+	else
+		cfid = e_contact_field_id_from_vcard (field_name);
+
+	for (ii = 0; ii < G_N_ELEMENTS (mappings); ii++) {
+		if (mappings[ii].field_id == cfid) {
+			return mappings[ii].mapi_id;
+		}
+	}
+
+	return MAPI_E_RESERVED;
+}
+
+static ESExpResult *
+func_eval_text_compare (struct _ESExp *f,
+			gint argc,
+			struct _ESExpResult **argv,
+			gpointer user_data,
+			uint32_t fuzzy)
+{
+	struct EMapiSExpParserData *esp = user_data;
+	ESExpResult *r;
+
+	r = e_sexp_result_new (f, ESEXP_RES_INT);
+	r->value.number = -1;
+
+	if (argc == 2
+	    && argv[0]->type == ESEXP_RES_STRING
+	    && argv[1]->type == ESEXP_RES_STRING) {
+		const gchar *propname = argv[0]->value.string;
+		const gchar *propvalue = argv[1]->value.string;
+
+		if (propname && propvalue && g_ascii_strcasecmp (propname, "x-evolution-any-field") != 0) {
+			uint32_t proptag = get_proptag_from_field_name (propname, TRUE);
+
+			if (proptag != MAPI_E_RESERVED && ((proptag & 0xFFFF) == PT_UNICODE || (proptag & 0xFFFF) == PT_STRING8)) {
+				struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+				g_return_val_if_fail (res != NULL, NULL);
+
+				res->rt = RES_CONTENT;
+				res->res.resContent.fuzzy = fuzzy | FL_IGNORECASE;
+				res->res.resContent.ulPropTag = proptag;
+				res->res.resContent.lpProp.ulPropTag = proptag;
+				res->res.resContent.lpProp.value.lpszW = talloc_strdup (esp->mem_ctx, propvalue);
+
+				g_ptr_array_add (esp->res_parts, res);
+				r->value.number = esp->res_parts->len - 1;
+			} else if (g_ascii_strcasecmp (propname, "email") == 0) {
+				struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+				g_return_val_if_fail (res != NULL, NULL);
+
+				res->rt = RES_OR;
+				res->res.resOr.cRes = 3;
+				res->res.resOr.res = talloc_zero_array (esp->mem_ctx, struct mapi_SRestriction_or, res->res.resOr.cRes + 1);
+
+				proptag = get_proptag_from_field_name ("email_1", TRUE);
+				res->res.resOr.res[0].rt = RES_CONTENT;
+				res->res.resOr.res[0].res.resContent.fuzzy = fuzzy | FL_IGNORECASE;
+				res->res.resOr.res[0].res.resContent.ulPropTag = proptag;
+				res->res.resOr.res[0].res.resContent.lpProp.ulPropTag = proptag;
+				res->res.resOr.res[0].res.resContent.lpProp.value.lpszW = talloc_strdup (esp->mem_ctx, propvalue);
+
+				proptag = get_proptag_from_field_name ("email_2", TRUE);
+				res->res.resOr.res[1].rt = RES_CONTENT;
+				res->res.resOr.res[1].res.resContent.fuzzy = fuzzy | FL_IGNORECASE;
+				res->res.resOr.res[1].res.resContent.ulPropTag = proptag;
+				res->res.resOr.res[1].res.resContent.lpProp.ulPropTag = proptag;
+				res->res.resOr.res[1].res.resContent.lpProp.value.lpszW = talloc_strdup (esp->mem_ctx, propvalue);
+
+				proptag = get_proptag_from_field_name ("email_3", TRUE);
+				res->res.resOr.res[2].rt = RES_CONTENT;
+				res->res.resOr.res[2].res.resContent.fuzzy = fuzzy | FL_IGNORECASE;
+				res->res.resOr.res[2].res.resContent.ulPropTag = proptag;
+				res->res.resOr.res[2].res.resContent.lpProp.ulPropTag = proptag;
+				res->res.resOr.res[2].res.resContent.lpProp.value.lpszW = talloc_strdup (esp->mem_ctx, propvalue);
+
+				g_ptr_array_add (esp->res_parts, res);
+				r->value.number = esp->res_parts->len - 1;
+			}
+		}
+	}
+
+	return r;
+}
+
+static ESExpResult *
+func_eval_contains (struct _ESExp *f,
+		    gint argc,
+		    struct _ESExpResult **argv,
+		    gpointer user_data)
+{
+	return func_eval_text_compare (f, argc, argv, user_data, FL_SUBSTRING);
+}
+
+static ESExpResult *
+func_eval_is (struct _ESExp *f,
+	      gint argc,
+	      struct _ESExpResult **argv,
+	      gpointer user_data)
+{
+	return func_eval_text_compare (f, argc, argv, user_data, FL_FULLSTRING);
+}
+static ESExpResult *
+func_eval_beginswith (struct _ESExp *f,
+		      gint argc,
+		      struct _ESExpResult **argv,
+		      gpointer user_data)
+{
+	return func_eval_text_compare (f, argc, argv, user_data, FL_PREFIX);
+}
+static ESExpResult *
+func_eval_endswith (struct _ESExp *f,
+		    gint argc,
+		    struct _ESExpResult **argv,
+		    gpointer user_data)
+{
+	/* no suffix, thus at least substring is used */
+	return func_eval_text_compare (f, argc, argv, user_data, FL_SUBSTRING);
+}
+
+static ESExpResult *
+func_eval_field_exists (struct _ESExp *f,
+			gint argc,
+			struct _ESExpResult **argv,
+			gpointer user_data,
+			gboolean is_contact_field)
+{
+	struct EMapiSExpParserData *esp = user_data;
+	ESExpResult *r;
+
+	r = e_sexp_result_new (f, ESEXP_RES_INT);
+	r->value.number = -1;
+
+	if (argc == 1 && argv[0]->type == ESEXP_RES_STRING) {
+		const gchar *propname = argv[0]->value.string;
+		uint32_t proptag = get_proptag_from_field_name (propname, is_contact_field);
+
+		if (proptag != MAPI_E_RESERVED) {
+			struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+			g_return_val_if_fail (res != NULL, NULL);
+
+			res->rt = RES_EXIST;
+			res->res.resExist.ulPropTag = proptag;
+
+			g_ptr_array_add (esp->res_parts, res);
+			r->value.number = esp->res_parts->len - 1;
+		} else if (g_ascii_strcasecmp (propname, "email") == 0) {
+			struct mapi_SRestriction *res = talloc_zero (esp->mem_ctx, struct mapi_SRestriction);
+			g_return_val_if_fail (res != NULL, NULL);
+
+			res->rt = RES_OR;
+			res->res.resOr.cRes = 3;
+			res->res.resOr.res = talloc_zero_array (esp->mem_ctx, struct mapi_SRestriction_or, res->res.resOr.cRes + 1);
+
+			res->res.resOr.res[0].rt = RES_EXIST;
+			res->res.resOr.res[0].res.resExist.ulPropTag = get_proptag_from_field_name ("email_1", TRUE);
+
+			res->res.resOr.res[1].rt = RES_EXIST;
+			res->res.resOr.res[1].res.resExist.ulPropTag = get_proptag_from_field_name ("email_2", TRUE);
+
+			res->res.resOr.res[2].rt = RES_EXIST;
+			res->res.resOr.res[2].res.resExist.ulPropTag = get_proptag_from_field_name ("email_3", TRUE);
+
+			g_ptr_array_add (esp->res_parts, res);
+			r->value.number = esp->res_parts->len - 1;
+		}
+	}
+
+	return r;
+}
+
+static ESExpResult *
+func_eval_exists (struct _ESExp *f,
+		  gint argc,
+		  struct _ESExpResult **argv,
+		  gpointer user_data)
+{
+	return func_eval_field_exists (f, argc, argv, user_data, TRUE);
+}
+
+static ESExpResult *
+func_eval_exists_vcard (struct _ESExp *f,
+			gint argc,
+			struct _ESExpResult **argv,
+			gpointer user_data)
+{
+	return func_eval_field_exists (f, argc, argv, user_data, FALSE);
+}
+
+struct mapi_SRestriction *
+mapi_book_utils_sexp_to_restriction (TALLOC_CTX *mem_ctx, const gchar *sexp_query)
+{
+	/* 'builtin' functions */
+	static const struct {
+		const gchar *name;
+		ESExpFunc *func;
+		gint type;		/* set to 1 if a function can perform shortcut evaluation, or
+					   doesn't execute everything, 0 otherwise */
+	} check_symbols[] = {
+		{ "and", 		term_eval_and,		0 },
+		{ "or", 		term_eval_or,		0 },
+		{ "not", 		term_eval_not,		0 },
+
+		{ "contains",		func_eval_contains,	0 },
+		{ "is",			func_eval_is,		0 },
+		{ "beginswith",		func_eval_beginswith,	0 },
+		{ "endswith",		func_eval_endswith,	0 },
+		{ "exists",		func_eval_exists,	0 },
+		{ "exists_vcard",	func_eval_exists_vcard,	0 }
+	};
+
+	gint i;
+	ESExp *sexp;
+	ESExpResult *r;
+	struct EMapiSExpParserData esp;
+	struct mapi_SRestriction *restriction;
+
+	g_return_val_if_fail (sexp_query != NULL, NULL);
+
+	esp.mem_ctx = mem_ctx;
+	sexp = e_sexp_new ();
+
+	for (i = 0; i < G_N_ELEMENTS (check_symbols); i++) {
+		if (check_symbols[i].type == 1) {
+			e_sexp_add_ifunction (sexp, 0, check_symbols[i].name,
+					      (ESExpIFunc *) check_symbols[i].func, &esp);
+		} else {
+			e_sexp_add_function (sexp, 0, check_symbols[i].name,
+					     check_symbols[i].func, &esp);
+		}
+	}
+
+	e_sexp_input_text (sexp, sexp_query, strlen (sexp_query));
+	if (e_sexp_parse (sexp) == -1) {
+		e_sexp_unref (sexp);
+		return NULL;
+	}
+
+	esp.res_parts = g_ptr_array_new ();
+	r = e_sexp_eval (sexp);
+
+	restriction = NULL;
+	if (r && r->type == ESEXP_RES_INT && r->value.number >= 0 && r->value.number < esp.res_parts->len)
+		restriction = g_ptr_array_index (esp.res_parts, r->value.number);
+
+	e_sexp_result_free (sexp, r);
+
+	e_sexp_unref (sexp);
+	g_ptr_array_free (esp.res_parts, TRUE);
+
+	return restriction;
+}
